/* xmstr.c - xmetoo string builtin functions */

#include <stdio.h>
#include "xmetoo.h"


/* external variables */
extern struct node *xlstack ;


/* external procedures */
extern char *strcpy() ;
extern char *strcat() ;
extern int atoi() ;
extern struct node* xlenter() ;
extern int issym() ;

/*. xstrlen - builtin function strlen .*/
/*. strlen : expstr... -> int .*/
static struct node *xstrlen( args )
struct node *args ;
{
    struct node arg ;
    struct node *oldstk, *val ;
    int total ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, NULL ) ;

    /* initialize */
    arg.n_ptr = args ;
    total = 0 ;

    /* loop over args and total */
    while( arg.n_ptr != NULL )
        total += strlen( xlevmatch( STR, &arg.n_ptr )->n_str ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* create the value node */
    val = newnode( INT ) ;
    val->n_int = total ;

    /* return the total */
    return( val ) ;
}


/*. xstrcat - builtin function strcat .*/
/*. strcat : expstr... -> str .*/
static struct node *xstrcat( args )
struct node *args ;
{
    struct node arg, val, rval ;
    struct node *oldstk ;
    int newlen ;
    char *result, *argstr, *newstr ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &val, &rval, NULL ) ;

    /* initialize */
    arg.n_ptr = args ;
    rval.n_ptr = newnode( STR ) ;
    rval.n_ptr->n_str = result = xmstralloc( 0 );
    *result = 0 ;

    /* loop over args */
    while( arg.n_ptr != NULL ) {

        /* get next argument */
        val.n_ptr = xlevmatch( STR, &arg.n_ptr ) ;
        argstr = val.n_ptr->n_str ;

        /* compute length of result */
        if( (newlen = strlen( result ) + strlen( argstr )) > STRMAX )
            xlfail( "string too long" ) ;

        /* allocate string and copy */
        newstr = xmstralloc( newlen ) ;
        strcpy( newstr, result) ;
        strfree( result ) ;
        rval.n_ptr->n_str = result = strcat( newstr, argstr ) ; }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the new string */
    return( rval.n_ptr ) ;
}


/*. substr - builtin function substr .*/
/*. substr : expstr expint [expint] -> str .*/
static struct node *substr( args )
struct node *args ;
{
    struct node arg, src, val ;
    struct node *oldstk ;
    int start, forlen, srclen ;
    char *srcptr, *dstptr ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &src, &val, NULL) ;

    /* initialize */
    arg.n_ptr = args ;
    
    /* get string and its length */
    src.n_ptr = xlevmatch( STR, &arg.n_ptr ) ;
    srcptr = src.n_ptr->n_str ;
    srclen = strlen (srcptr) ;

    /* get starting pos -- must be present */
    start = xlevmatch( INT, &arg.n_ptr )->n_int ;

    /* get length -- if not present use remainder of string */
    if( arg.n_ptr != NULL )
        forlen = xlevmatch( INT, &arg.n_ptr )->n_int ;
    else
        forlen = srclen  ;

    /* make sure there aren't any more arguments */
    xllastarg( arg.n_ptr ) ;

    /* if start beyond string -- return null string */
    if( (start > srclen) || (start < 1) ) {
        start = 1 ;
        forlen = 0 ; }
    if( forlen < 0 ) forlen = 0 ;
	
    /* don't take more than exists */
    if( start + forlen > srclen )
        forlen = srclen - start + 1 ;

    /* create return node */
    val.n_ptr = newnode( STR ) ;
    val.n_ptr->n_str = dstptr = xmstralloc( forlen ) ;

    /* move string */
    for( srcptr += start-1 ; forlen-- ; *dstptr++ = *srcptr++) ;
    *dstptr = 0 ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the substring */
    return( val.n_ptr ) ;
}


/*. ascii - builtin function ascii .*/
/*. ascii : expstr -> int .*/
static struct node *ascii( args )
struct node *args ;
{
    struct node *val ;
    int c ;

    c = *(xlevmatch( STR, &args )->n_str) ;

    /* make sure there aren't any more arguments */
    xllastarg(args);

    /* build return node */
    val = newnode( INT ) ;
    val->n_int = c ;

    /* return the character */
    return( val ) ;
}


/*. chr - builtin function chr .*/
/*. chr : expint -> str .*/
static struct node *chr( args )
struct node *args ;
{
    struct node val ;
    struct node *oldstk ;
    char *sptr ;

    /* create a new stack frame */
    oldstk = xlsave( &val, NULL ) ;

    /* build return node */
    val.n_ptr = newnode( STR ) ;
    val.n_ptr->n_str = sptr = xmstralloc( 1 ) ;
    *sptr = xlevmatch( INT, &args )->n_int ;
    *++sptr = 0 ;

    /* make sure there aren't any more arguments */
    xllastarg( args ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the new string */
    return( val.n_ptr ) ;
}


/*. xatoi - builtin function atoi .*/
/*. atoi : expstr -> int .*/
static struct node *xatoi( args )
struct node *args ;
{
    struct node *val ;
    int n ;

    /* get the string and convert it */
    n = atoi( xlevmatch( STR, &args )->n_str ) ;

    /* make sure there aren't any more arguments */
    xllastarg( args ) ;

    /* create the value node */
    val = newnode( INT ) ;
    val->n_int = n ;

    /* return the number */
    return( val ) ;
}


/*. xitoa - builtin function itoa .*/
/*. itoa : expint -> str .*/
static struct node *xitoa( args )
struct node *args ;
{
    struct node *val ;
    char buf[20] ;

    /* get the integer and convert it */
    sprintf( buf, "%d", xlevmatch( INT, &args )->n_int ) ;

    /* make sure there aren't any more arguments */
    xllastarg( args ) ;

    /* create the value node */
    val = newnode( STR ) ;
    val->n_str = strsave( buf ) ;

    /* return the string */
    return( val ) ;
}


/*. symstr - builtin function symstr .*/
/*. symstr : simb -> str .*/
static struct node *symstr( args )
struct node *args ;
{
    struct node arg;
    struct node *oldstk, *val ;
      
    /* create a new stack frame */
    oldstk = xlsave( &arg, NULL) ;             

    /* get the symbol */
     arg.n_ptr = xlevmatch( SYM, &args ) ;

    /* make sure there aren't any more arguments */
    xllastarg( args ) ;

    /* create the value node */
    val = newnode( STR ) ;
    val->n_str = strsave(arg.n_ptr->n_symname) ;

    /* restore the previous stack frame */
    xlstack = oldstk;

    /* return the string */
    return( val ) ;
}


/*. strsym - builtin function strsym .*/
/*. strsym : str -> simb .*/
static struct node *strsym( args )
struct node *args ;
{
    struct node *arg, *val ;
    char *symname ;

    /* get the string */
    arg = xlevmatch( STR, &args ) ;
    
    /* make sure there aren't any more arguments */
    xllastarg( args ) ;

    /* verify if the string is a valid symbol name */
    symname = arg->n_str;
    while( issym(*(symname++)) ) ;
   
 /* create the value node if possible */
    if( *(--symname) == '\0' )
         val = xlenter( arg->n_str ) ;
    else xlfail( "conversion not possible" );

    return( val ) ;
}


/*. min - builtin function min .*/
/*. min : expstr expstr... --> str .*/
/*. (implemented by "min" in integer algebra module).*/


/*. max - builtin function max .*/
/*. max : expstr expstr... --> str .*/
/*. (implemented by "max" in integer algebra module).*/


/*. xmstrinit - xmetoo string initialization routine .*/
void xmstrinit()
{
    xlsubr( "strlen", xstrlen ) ;
    xlsubr( "strcat", xstrcat ) ;
    xlsubr( "substr", substr ) ;
    xlsubr( "ascii", ascii ) ;
    xlsubr( "chr", chr ) ;
    xlsubr( "atoi", xatoi ) ;
    xlsubr( "itoa", xitoa ) ;
    xlsubr( "symstr", symstr) ;
    xlsubr( "strsym", strsym);
}

