/* xmxlisp.c - xlisp imported builtin functions */

#include <stdio.h>
#include "xmetoo.h"


/* external variables */
extern struct node *xlstack ;
extern struct node *true ;
extern struct node *false ;
#ifdef USERTYPE
extern struct t_table *name_table[] ;
#endif


/* external procedures */
extern void exit() ;
extern int system() ;
extern char *strcat() ;
extern void subst_name() ;


/* local variables */
static struct node *a_subr ;
static struct node *a_list ;
static struct node *a_sym ;
static struct node *a_int ;
static struct node *a_str ;
static struct node *a_fptr ;
#ifdef COMM
static struct node *a_chan ;
#endif
#ifdef USERTYPE
static struct node *a_type ;
static struct node *a_label ;
static struct node *a_mactype ;
#endif
static struct node *a_set ;
static struct node *a_rel ;
static struct node *a_ff ;
static struct node *a_fun ;
#ifdef LAZY
static struct node *a_rec ;
#endif


/*. progn - builtin function progn .*/
/*. progn : exp... -> val .*/
static struct node *progn( args )
struct node *args ;
{
    struct node arg ;
    struct node *oldstk, *val ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, NULL ) ;

    arg.n_ptr = args ;
    val = NULL;
    while( arg.n_ptr != NULL )
        val = xlevarg( &arg.n_ptr ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the value of the last expression */
    return( val ) ;
}


/*. prog1 - builtin function prog1 .*/
/*. prog1 : exp... -> value .*/
static struct node *prog1( args )
struct node *args ;
{
    struct node arg, val ;
    struct node *oldstk ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &val, NULL ) ;

    arg.n_ptr = args ;
    if( arg.n_ptr != NULL )
        val.n_ptr = xlevarg( &arg.n_ptr ) ;
    while( arg.n_ptr != NULL )
        xlevarg( &arg.n_ptr ) ;

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return pointer to the value of the first expression */
    return( val.n_ptr ) ;
}


/*. fwhile - builtin function while .*/
/*. while : expbool exp... -> val .*/
static struct node *fwhile( args )
struct node *args ;
{
    struct node farg ;
    struct node *oldstk, *val, *b ;

    /* create a new stack frame */
    oldstk = xlsave( &farg, NULL ) ;

    /* initialize */
    farg.n_ptr = args ;

    /* loop until test fails */
    val = NULL ;
    if( ((b = xlevarg( &args )) != true) && (b != false) )
        xlfail( "bad argument type" ) ;
    while( b == true ) {
        while( args != NULL )
            val = xlevarg( &args ) ;
        args = farg.n_ptr ;
        if( ((b = xlevarg( &args )) != true) && (b != false) )
            xlfail( "bad argument type" ) ; }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the value of the last evaluated expression */
    return( val ) ;
}


/*. frepeat - builtin function repeat .*/
/*. repeat : expint exp... -> val .*/
static struct node *frepeat( args )
struct node *args ;
{
    struct node farg ;
    struct node *oldstk, *val ;
    int cnt ;

    /* create a new stack frame */
    oldstk = xlsave( &farg, NULL) ;

    /* initialize */
    farg.n_ptr = args ;

    /* evaluate the repeat count */
    cnt = xlevmatch( INT, &args )->n_int ;

    /* save the first expression to repeat */
    farg.n_ptr = args ;

    /* loop until test fails */
    for( val = NULL ; cnt > 0 ; cnt--, args = farg.n_ptr ) {

        /* evaluate each remaining argument */
        while( args != NULL )
            val = xlevarg( &args ) ; }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* return the last test expression value */
    return( val ) ;
}


/*. foreach - builtin function foreach .*/
/*. foreach : sym explist exp... -> val .*/
static struct node *foreach( args )
struct node *args ;
{
    struct node arg, sym, list, code, oldbnd, par_list ;
    struct node *oldstk, *val ;
    char name[STRMAX+1] ;

    /* create a new stack frame */
    oldstk = xlsave( &arg, &sym, &list, &code, &oldbnd, &par_list, NULL ) ;

    /* initialize */
    arg.n_ptr = args ;

    /* get the symbol to bind to each list element */
    sym.n_ptr = xlmatch( SYM, &arg.n_ptr ) ;

    /* get the list to iterate over */
    /*     list.n_ptr = xlevmatch( LIST, &arg.n_ptr ) ; */
    /* jj para permitir foreach de SET FF LIST */

    list.n_ptr = xlevarg(&arg.n_ptr);

#ifdef USERTYPE
    while( (list.n_ptr != NULL) && (list.n_ptr->n_type >= USERDEFTYPE) )
        list.n_ptr = list.n_ptr->n_symvalue ;
#endif /* USERTYPE jj */

 if( (list.n_ptr != NULL) && 
     (list.n_ptr->n_type != SET) && 
     (list.n_ptr->n_type != REL) && 
     (list.n_ptr->n_type != FF) && 
     (list.n_ptr->n_type != LIST) )
        xlfail( "bad argument type" ) ;


    /* save the pointer to the code */
    code.n_ptr = arg.n_ptr ;

    if( *(sym.n_ptr->n_symname) != ' ' ) {
        par_list.n_ptr = newnode( LIST ) ;
        par_list.n_ptr->n_listvalue = sym.n_ptr ;
        *name = ' ' ; *(name+1) = '\0' ;
        strcat( name, sym.n_ptr->n_symname ) ;
        sym.n_ptr = xlenter( name ) ;
        while( arg.n_ptr != NULL ) {
            subst_name( &par_list.n_ptr, &(arg.n_ptr->n_listvalue) ) ;
            arg.n_ptr = arg.n_ptr->n_listnext ; }
        arg.n_ptr = code.n_ptr ; }

    /* save the old binding of the symbol */
    oldbnd.n_ptr = sym.n_ptr->n_symvalue ;

    /* loop until test fails */
    val = NULL ;
    while( list.n_ptr != NULL ) {

        /* bind the symbol to the list element */
        sym.n_ptr->n_symvalue = list.n_ptr->n_listvalue ;

        /* evaluate each remaining argument */
        while( arg.n_ptr != NULL )
            val = xlevarg( &arg.n_ptr ) ;

        /* point to the next list element */
        list.n_ptr = list.n_ptr->n_listnext ;

        /* restore the pointer to the code */
        arg.n_ptr = code.n_ptr ; }

    /* restore the previous stack frame */
    xlstack = oldstk ;

    /* restore the old binding of the symbol */
    sym.n_ptr->n_symvalue = oldbnd.n_ptr ;

    /* return the last test expression value */
    return( val ) ;
}


/*. fexit - builtin function exit .*/
/*. exit : [expint] -> .*/
static struct node *fexit( args )
struct node *args ;
{
    int n ;

    if( args == NULL )
        n = 0 ;
    else {
        n = xlevmatch( INT, &args )->n_int ;
        xllastarg( args ) ; }

    exit( n ) ;
}


/*. m2system - builtin function system .*/
/*. system : [expstr] -> int .*/
static struct node *m2system( args )
struct node *args ;
{
    struct node *n ;
    char *s ;
    int r ;

    if( args == NULL )
#ifdef __TURBOC__
        s = "COMMAND" ;
#else
#ifdef unix
        s = "sh" ;
#else
        s = NULL ;
#endif
#endif
    else
        s = xlevmatch( STR, &args )->n_str ;
    xllastarg( args ) ;

    r = (s == NULL) ? -1 : system( s ) ;
    n = newnode( INT ) ;
    n->n_int = r ;
    return( n ) ;
}


/*. type - builtin function type .*/
/*. type : exp -> sym .*/
static struct node *type( args )
struct node *args ;
{
    struct node *arg ;

    arg = xljustevarg( &args ) ;
    xllastarg( args ) ;

    if( arg == NULL )
        return( NULL ) ;

    switch( arg->n_type ) {
    case SUBR :
        return( a_subr ) ;
    case LIST :
        return( a_list ) ;
    case SYM :
        return ( a_sym ) ;
    case INT :
        return( a_int ) ;
    case STR :
        return( a_str ) ;
    case FPTR :
        return( a_fptr ) ;
#ifdef COMM
    case CHAN :
        return( a_chan ) ;
#endif
#ifdef USERTYPE
    case TYPE:
        return( a_type ) ;
    case LABEL:
        return( a_label ) ;
    case MACTYPE:
        return( a_mactype ) ;
#endif
    case SET :
        return( a_set ) ;
    case REL :
        return( a_rel ) ;
    case FF :
        return( a_ff ) ;
    case FUN :
        return( a_fun ) ;
#ifdef LAZY
    case REC :
    case LISTREC :
        return( a_rec ) ;
#endif
    default :
#ifdef USERTYPE
       if( arg->n_type >= USERDEFTYPE )
           return( xlenter( name_table[arg->n_type - USERDEFTYPE]->tname ) ) ; 
#endif
        xlfail( "Bad node" ) ; }
}


/*. xmxlispinit - xlisp initialization routine .*/
void xmxlispinit()
{
    a_subr = xlenter( "SUBR" ) ;
    a_list = xlenter( "LIST" ) ;
    a_sym = xlenter( "SYM" ) ;
    a_int = xlenter( "INT" ) ;
    a_str = xlenter( "STR" ) ;
    a_fptr = xlenter( "FPTR" ) ;
#ifdef COMM
    a_chan = xlenter( "CHAN" ) ;
#endif
#ifdef USERTYPE
    a_type = xlenter( "TYPE" ) ;
    a_label = xlenter( "LABEL" ) ;
    a_mactype = xlenter( "MACTYPE" ) ;
#endif
    a_set = xlenter( "SET" ) ;
    a_rel = xlenter( "REL" ) ;
    a_ff = xlenter( "FF" ) ;
    a_fun = xlenter( "FUN" ) ;
#ifdef LAZY
    a_rec = xlenter( "delayed_expression" ) ;
#endif

    xlsubr( "while", fwhile ) ;
    xlsubr( "repeat", frepeat ) ;
    xlsubr( "progn", progn ) ;
    xlsubr( "prog1", prog1 ) ;
    xlsubr( "foreach", foreach ) ;
    xlsubr( "!", fexit ) ;
    xlsubr( "sh", m2system ) ;
    xlsubr( "type", type) ;
}

